import flatdict
import pymongo
from bson import DBRef

from asynclib.database.mongo import MongoDBAMQPProxy, MongoDBManager
from asynclib.utils.functions import get_from_dict


class DenormDBManager(MongoDBManager):

    def __init__(self, mongo):
        super(DenormDBManager, self).__init__('sys', 'denorm', mongo)

    async def create_indexes(self):
        await self.db.create_indexes([
            ('db_name', pymongo.ASCENDING),
            ('collection_name', pymongo.ASCENDING),
            ('ref_id', pymongo.ASCENDING)
        ], background=True, name='denorm')

    async def create(self, loader, callback):
        if loader.extra is not None:
            for _id in loader._id_list if hasattr(loader, '_id_list') else [loader._id]:
                data = {
                    '$set': {
                        'db_name': loader.db_name,
                        'collection_name': loader.collection_name,
                        'ref_id': _id
                    },
                    '$addToSet': {
                        'callback': {
                            'db_name': callback.db_name,
                            'collection_name': callback.collection_name,
                            'extra': loader.extra,
                            'query_path': callback.query_path,
                            'set_path': callback.set_path,
                            'query': callback.query
                        }
                    }
                }
                await self.db.update_one({
                    'db_name': loader.db_name,
                    'collection_name': loader.collection_name,
                    'ref_id': _id}, data, upsert=True)

    async def denorm_collection(self, db, old_map, new_map):
        for key, value in new_map.items():
            await self.denorm(db, value['_id'], old_map[key], value)

    async def denorm(self, db, _id, old, new):
        denorm_pack = await self.db.find_one({'db_name': db.db_name, 'collection_name': db.collection_name, 'ref_id': _id})
        if denorm_pack is None:
            return
        denorm_items = [i for i in denorm_pack.get('callback', []) if
                        (i.get('extra') is not None and i.get('db_name') is not None)]
        keys_for_check = self.get_keys_for_check(denorm_items)
        updated_keys = self.get_updated_keys(old, new, keys_for_check)
        for denorm_item in denorm_items:
            extra_keys = denorm_item.get('extra')
            if not self.any_in_array(extra_keys, updated_keys):
                continue
            item = new
            extra = self.prepare_extra(item, denorm_item.get('extra'))
            item = DBRef(
                db.collection_name, _id, db.db_name, _extra=extra
            )
            ref_db = MongoDBAMQPProxy(
                db_name=denorm_item.get('db_name'),
                collection_name=denorm_item.get('collection_name'),
            )
            query = {denorm_item.get('query_path', '') + '.$id': _id}
            if denorm_item.get('query'):
                query.update(denorm_item.get('query'))
            await ref_db.update_many_denormalized(
                query,
                {'$set': {denorm_item.get('set_path'): item}}
            )

    def get_keys_for_check(self, items):
        return set([keypath for item in items for keypath in item.get('extra')])

    def get_updated_keys(self, old, new, keys_for_check):
        updated_keys = []
        for key in keys_for_check:
            if get_from_dict(old, key) != get_from_dict(new, key):
                updated_keys.append(key)
        return updated_keys

    def any_in_array(self, array1, array2):
        return len(set(array2).intersection(set(array1))) > 0

    def prepare_extra(self, item, fields):
        extra = flatdict.FlatDict({}, delimiter='.')
        item = flatdict.FlatDict(item, delimiter='.')
        for key in fields:
            extra[key] = item.get(key)
        extra = extra.as_dict()
        return extra



import datetime

from asynclib.http.error import RequestParamsError, BaseError


class ValidationError(BaseError):

    def __init__(self, code):
        super(ValidationError, self).__init__(code=code, description='Validation error')


class ValidatorErrorCodes(object):
    VALUE_REQUIRED = 'VALUE_REQUIRED'
    VALUE_TYPE_ERROR = 'VALUE_TYPE_ERROR'
    ENUM_ERROR = 'ENUM_ERROR'


def raise_if(condition, code):
    if condition:
        raise ValidationError(code=code)


def raise_if_not_instance(value, cls):
    raise_if(not isinstance(value, cls), ValidatorErrorCodes.VALUE_TYPE_ERROR)


class Validator(object):

    async def validate(self, value):
        raise NotImplementedError()



class OneOf(Validator):

    def __init__(self, choices):
        self.choices = choices

    async def validate(self, value):
        raise_if(value not in self.choices, ValidatorErrorCodes.ENUM_ERROR)


class Field(object):

    def __init__(self, required=False, validators=None):
        self.required = required
        self.validators = validators if validators is not None else []

    async def deserialize(self, value):
        if self.required and value is None:
            raise ValidationError(ValidatorErrorCodes.VALUE_REQUIRED)
        value = await self._deserialize(value)
        self.validate(value)
        return value

    async def _deserialize(self, value):
        return value

    async def validate(self, value):
        for validator in self.validators:
            await validator.validate(value)


class DefaultTypeField(Field):

    def __init__(self, _type, required=False, validators=None):
        self._type = _type
        super(DefaultTypeField, self).__init__(required=required, validators=validators)

    async def _deserialize(self, value):
        if value is None:
            return None
        raise_if_not_instance(value, self._type)
        return value


class String(Field):

    async def _deserialize(self, value):
        if value is None:
            return None
        raise_if_not_instance(value, (bytes, str))
        if isinstance(value, bytes):
            value = value.decode('utf-8')
        await self.validate(value)
        return value


class Integer(DefaultTypeField):
    def __init__(self, required=False, validators=None):
        super(Integer, self).__init__(int, required=required, validators=validators)


class Float(DefaultTypeField):

    def __init__(self, required=False, validators=None):
        super(Float, self).__init__(float, required=required, validators=validators)


class Datetime(String):

    async def _deserialize(self, value):
        value = await super(Datetime, self)._deserialize(value)
        return datetime.datetime.strptime(value[:19], '%Y-%m-%dT%H:%M:%S')



async def parse_args(args, arg_map):
    """
    :type args: dict
    :type arg_map: dict
    """
    serialized = {}
    error_map = {}
    for key, field in arg_map.items():
        try:
            value = await field.deserialize(args.get(key))
            serialized[key] = value
        except BaseError as error:
            error_map[key] = error.code
    if error_map:
        raise RequestParamsError(error_map)
    return serialized





class BaseError(Exception):

    def __init__(self, code, description, status=None, body=None):
        self.body = body
        self.description = description
        self.code = code
        self.status = status

    def map(self):
        _map = {
            'description': self.description,
            'code': self.code
        }
        if self.body is not None:
            _map['body'] = self.body
        if self.status is not None:
            _map['status'] = self.status
        return _map

    @classmethod
    def init_with(cls, attributes):
        return cls(
            code=attributes.get('code'),
            description=attributes.get('description'),
            status=attributes.get('status'),
            body=attributes.get('body')
        )


class RequestParamsError(BaseError):

    def __init__(self, error_map):
        super(RequestParamsError, self).__init__(
            status=400, code='REQUEST_PARAMS_ERROR', description='Request send failed parameters', body=error_map
        )


from random import randint


def filter_only_keys(dictionary, keys):
    filtered = {}
    for k in keys:
        value = dictionary.get(k)
        if value is not None:
            filtered[k] = value
    return filtered


def random_with_N_digits(n):
    range_start = 10 ** (n - 1)
    range_end = (10 ** n) - 1
    return randint(range_start, range_end)


def get_from_dict(data, path):
    path = path.split('.')
    if len(path) == 1:
        try:
            return data.get(path[0])
        except AttributeError:
            try:
                return getattr(data, path[0])
            except AttributeError:
                return None
    else:
        try:
            data = data.get(path[0])
        except AttributeError:
            data = getattr(data, path[0])
        return get_from_dict(data, '.'.join(path[1:]))





import asyncio
from collections import defaultdict
import motor.motor_asyncio
import pymongo
from bson import DBRef
from pymongo import errors

from asynclib.amqp.client import AMQPClient


class Mongo(object):

    def __init__(self, url, name):
        self._client = None
        self.name = name
        self.url = url

    def get_collection(self, collection_name, read_preference):
        if self._client is None:
            self._client = motor.motor_asyncio.AsyncIOMotorClient(self.url)
        return self._client.get_database(self.name, read_preference=read_preference).get_collection(collection_name)


def auto_reconnect(func):
    async def wrapper(*args, **kwargs):
        for attempt in range(5):
            try:
                return await func(*args, **kwargs)
            except errors.AutoReconnect:
                print("reconnecting to db")
                wait_t = 0.5 * pow(2, attempt)
                asyncio.sleep(wait_t)

    return wrapper


class MongoDBAMQPProxy(object):

    def __init__(self, db_name, collection_name):
        self.db_name = db_name
        self.collection_name = collection_name
        self.amqp_client = AMQPClient(name=db_name)

    def __getattr__(self, item):
        async def query_maker(*args, **kwargs):
            await self.amqp_client.connect()
            result = await self.amqp_client.call('mongo_query', parameters={'method': item,
                                                                            'collection': self.collection_name
                ,'args': args, 'kwargs': kwargs})
            await self.amqp_client.close()
            return result
        return query_maker


class DenormDBProxy(MongoDBAMQPProxy):

    def __init__(self):
        super(DenormDBProxy, self).__init__('denorm', 'ref_map')


class MongoDBManager(object):

    def __init__(self, collection_name, mongo_retriever,
                 read_preference=pymongo.ReadPreference.SECONDARY_PREFERRED):
        self.mongo_retriever = mongo_retriever
        self.read_preference = read_preference
        self.collection_name = collection_name
        self._db = None

    @property
    def db(self) -> motor.motor_asyncio.AsyncIOMotorCollection:
        if self._db is None:
            self._db = self.mongo_retriever.mongo.get_collection(self.collection_name, self.read_preference)
        return self._db

    def __getattr__(self, item):
        #proxy for query
        return getattr(self.db, item)

    @auto_reconnect
    async def find_by_id_list(self, id_list):
        return await self.db.find({'$id': {'$in': id_list}})

    @auto_reconnect
    async def get_by_id(self, _id):
        return await self.db.find_one({'_id': _id})

    @auto_reconnect
    async def _create(self, parameters):
        result = await self.db.insert_one(parameters)
        parameters['_id'] = result.inserted_id
        return parameters

    async def cursor_map(self, cursor):
        result = await cursor.to_list()
        return {str(i['_id']): i for i in result}

    @auto_reconnect
    async def update_many_denormalized(self, query, data, *args, **kwargs):
        old_cursor_map = await self.cursor_map(self.db.find(query))
        result = await self.db.update_many(query, data, *args, **kwargs)
        if result.modified_count > 0:
            new_cursor = self.db.find(query)
            if await new_cursor.count() > 0:
                new_cursor_map = await self.cursor_map(new_cursor)
                denorm_db = DenormDBProxy()
                denorm_db.denorm_collection(self, old_cursor_map, new_cursor_map)
        return result

    @auto_reconnect
    async def update_one_denormalized(self, query, data, *args, **kwargs):
        doc_before = await self.db.find_one(query)
        doc_after = await self.db.find_one_and_update(
            query, data, return_document=pymongo.ReturnDocument.AFTER, *args, **kwargs
        )
        if doc_after:
            denorm_db = DenormDBProxy()
            denorm_db.denorm(self, doc_after['_id'], doc_before, doc_after)


async def cursor_to_result(cursor, skip=None, limit=None):
    """
    :param limit: int
    :param skip: int
    :type cursor: pymongo.Cursor
    """
    try:
        if limit is not None:
            limit = int(limit)
        if skip is not None:
            skip = int(skip)
    except (ValueError, TypeError):
        limit = None
        skip = None
    total = await cursor.count()
    if skip is not None:
        cursor.skip(skip)
    if limit is not None:
        cursor.limit(limit)
    return {
        'objects': await cursor.to_list(None),
        'total': total
    }


def collect_db_ref(items):
    values = []
    if isinstance(items, dict):
        items = items.values()
    for item in items:
        if isinstance(item, DBRef):
            values.append(DBRef(item.collection, item.id, item.database))
            values += collect_db_ref(item._DBRef__kwargs)
        elif isinstance(item, (list, dict)):
            values += collect_db_ref(item)
    return values


def append_db_ref(item, data):
    if isinstance(item, DBRef):
        item._DBRef__kwargs = append_db_ref(item._DBRef__kwargs, data)
        return data[item.database][item.collection].get(item.id) or item
    elif isinstance(item, list):
        return [append_db_ref(i, data) for i in item]
    elif isinstance(item, dict):
        return {k: append_db_ref(i, data) for k, i in item.items()}
    return item


async def deref(data, items):
    refs = defaultdict(list)
    for r in set(collect_db_ref(data)):
        if r.database + '.' + r.collection in items:
            refs[(r.database, r.collection)].append(r.id)
    docs = defaultdict(lambda: defaultdict(dict))
    for key, _id in refs.items():
        docs[key[0]][key[1]] = {i['_id']: i for i in await MongoDBAMQPProxy(key[0], key[1]).find({'_id': {'$in': _id}})}
    return append_db_ref(data, docs)






import json

import aiohttp.web

from asynclib.http.marshall import error_middleware


class RestServer(object):
    def __init__(self, name):
        self.name = name
        self.app = aiohttp.web.Application(middlewares=[error_middleware])
        self.app.router.add_get('/{}/heartbeat'.format(self.name), self.heartbeat)

    def add_resource(self, path, resource):
        for method in ['get', 'put', 'post', 'delete', 'patch']:
            if hasattr(resource, method):
                self.app.router.add_route(method.upper(), '/{}{}'.format(self.name, path), getattr(resource, method))

    def add_handler(self, path, method, handler):
        self.app.router.add_route(method, '/{}{}'.format(self.name, path), handler)

    async def heartbeat(self, request):
        return aiohttp.web.Response(body=json.dumps({'alive': True}), status=200, content_type='application/json')

    def run(self, host, port, loop):
        aiohttp.web.run_app(self.app, host=host, port=port, loop=loop)



import asyncio
import json

import re
import uuid

import aioamqp
from asynclib.amqp import logger
from asynclib.http.error import BaseError


class AMQPClient(object):
    def __init__(self, name='', prefix='am', host='localhost', user='guest', password='guest',
                 vhost='/', exchange_type='topic', dumper=json, prefetch_count=1):
        if not prefix.endswith('_'):
            prefix += '_'
        if len(name) > 0 and not name.endswith('_'):
            name += '_'
        self.credentials = {
            'host': host,
            'login': user,
            'password': password,
            'virtual_host': vhost
        }
        self.dumper = dumper
        self.prefix = prefix
        self.prefetch_count = prefetch_count
        self.name = name
        self.exchange_name = self.prefix + 'exchange_' + exchange_type
        self.queue_name = self.prefix + name + 'queue_' + exchange_type
        self.transport = None
        self.protocol = None
        self.channel = None
        self.exchange_type = exchange_type
        self.waiter = asyncio.Event()
        self._reply_queue = None

    async def connect(self):
        self.transport, self.protocol = await aioamqp.connect(**self.credentials)
        self.channel = await self.protocol.channel()

    async def __aenter__(self):
        await self.connect()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def __await__(self):
        return self.__aenter__().__await__()

    async def close(self):
        await self.protocol.close()
        self.transport.close()

    async def get_reply_queue(self):
        if self._reply_queue is None:
            self._reply_queue = (await self.channel.queue_declare(exclusive=True))['queue']
        return self._reply_queue

    async def on_reply(self, channel, body, envelope, properties):
        if properties.correlation_id == self.correlation_id:
            self.response = self.dumper.loads(body)
            self.waiter.set()

    async def call(self, key, **kwargs):
        self.correlation_id = str(uuid.uuid4())
        reply = await self.get_reply_queue()
        routing_key = (self.prefix + self.name).replace('_', '.') + key
        body = self.dumper.dumps({'parameters': kwargs})
        await self.channel.basic_consume(self.on_reply, queue_name=reply)
        await self.channel.basic_publish(body,
            self.exchange_name, routing_key=routing_key, properties={
                'correlation_id': self.correlation_id,
                'reply_to': reply
            }
        )
        await self.waiter.wait()
        if self.response.get('error') is not None:
            raise BaseError.init_with(self.response['error'])
        return self.response['result']


    async def publish(self, _name, *args, **kwargs):
        routing_key = (self.prefix + self.name).replace('_', '.') + _name
        body = self.dumper.dumps({'args': args, 'kwargs': kwargs})
        self.channel.basic_publish(body, self.exchange_name, routing_key=routing_key)
        logger.debug('Message %s with routing key %s published', body, routing_key)




import importlib
import os

import six

from asynclib.utils.singleton import MetaSingleton


@six.add_metaclass(MetaSingleton)
class Config(dict):
    def __init__(self, **kwargs):
        super(Config, self).__init__(**kwargs)
        self.reload()

    def reload(self):
        self._load_from_module()
        self._load_from_env()

    def _load_from_module(self):
        try:
            module = importlib.import_module(os.environ.get('CONFIG_MODULE', 'settings'))
            for key in dir(module):
                if key.isupper():
                    self[key] = getattr(module, key)
        except ImportError:
            pass

    def _load_from_env(self):
        for key, value in os.environ.items():
            if key.isupper():
                self[key] = value

    def get_as_int(self, key, default=None):
        if key not in self:
            return default
        if isinstance(self[key], int):
            return self[key]
        try:
            value = int(self[key])
            self[key] = value
            return value
        except ValueError:
            return default

    def get_as_list(self, key, default=None):
        if key not in self:
            return default
        if isinstance(self[key], list):
            return self[key]
        if isinstance(self[key], six.string_types):
            value = list(six.moves.urllib.parse.parse_qs(self[key], keep_blank_values=True).keys())
            self[key] = value
            return value
        return list(self[key])

    def get_as_bool(self, key, default=None):
        if key not in self:
            return default
        if isinstance(self[key], bool):
            return self[key]
        if isinstance(self[key], six.string_types):
            value = self[key].lower() in ('1', 'true', 'yes')
            self[key] = value
            return value
        return bool(self[key])

    def get_as_dict(self, key, default=None):
        if key not in self:
            return default
        if isinstance(self[key], dict):
            return self[key]
        if isinstance(self[key], six.string_types):
            value = {a: b[0] if len(b) == 1 else b for a, b in six.moves.urllib.parse.parse_qs(self[key]).items()}
            self[key] = value
            return value
        return dict(self[key])


from asynclib.denorm_s.db import DenormDBManager
from asynclib.service import Service, ServiceConfig


class DenormService(Service):

    def __init__(self, mongo_url):
        super(DenormService, self).__init__(ServiceConfig('denorm', mongo_url))

    def get_query_db(self, *args, **kwargs):
        return DenormDBManager(self.mongo)

    def run(self):
        self.run_private_api()


import json


class RedisObjectCache(object):

    def __init__(self, redis, object_prefix):
        self.redis = redis
        self.object_prefix = object_prefix

    async def set(self, obj, _id):
        await self.redis.execute('set', '{}{}.'.format(self.object_prefix, _id), json.dumps(obj))

    async def get(self, _id):
         return json.loads(await self.redis.execute('get', '{}{}.'.format(self.object_prefix, _id)))


from aioredis import create_pool

from asynclib.amqp.server import AMQPServer
from asynclib.database.mongo import Mongo, MongoDBManager
from asynclib.http.rest import RestServer
from asynclib.utils.loop import get_event_loop


class ServiceConfig(object):
    def __init__(self, name, mongo_url='mongo://127.0.0.1', redis_host='127.0.0.1',
                 amqp_host='127.0.0.1', amqp_user='guest', amqp_password='guest'):
        self.redis_host = redis_host
        self.name = name
        self.amqp_password = amqp_password
        self.amqp_user = amqp_user
        self.amqp_host = amqp_host
        self.mongo_url = mongo_url


class Service(object):

    def __init__(self, config: ServiceConfig):
        self.config = config
        self.private_api = AMQPServer(name=self.config.name, host=self.config.amqp_host,
                                      user=self.config.amqp_user, password=self.config.amqp_password)
        self.public_api = RestServer(name=self.config.name)
        self.loop = get_event_loop()
        self._mongo = None
        self._redis = None

    @property
    def mongo(self):
        if self._mongo is None:
            self._mongo = Mongo(url=self.config.mongo_url, name=self.config.name)
        return self._mongo

    @property
    def redis(self):
        if self._redis is None:
            self._redis = create_pool((self.config.redis_host, 6379), minsize=5, maxsize=10, loop=self.loop)
        return self._redis

    def get_query_db(self, *args, **kwargs):
        return MongoDBManager(args[0], self.mongo)

    async def mongo_query(self, parameters):
        db = self.get_query_db(parameters['collection'])
        coroutine = getattr(db, parameters['method'])
        return await coroutine(*parameters.get('args', []), **parameters.get('kwargs', {}))

    def run_public_api(self, host, port):
        self.public_api.run(host=host, port=port, loop=self.loop)

    def run_private_api(self):
        self.private_api.register_endpoint('mongo', self.mongo_query)
        self.private_api.run(self.loop)



# coding=utf-8
import bson

from asynclib.database.mongo import DenormDBProxy, MongoDBAMQPProxy


class RefLoader(object):
    def __init__(self, db_name, collection_name, _id, extra=None, item=None):
        self.db_name = db_name
        self.collection_name = collection_name
        self._id = _id
        self.extra = extra
        self.item = item


class RefListLoader(object):
    def __init__(self, db_name, collection_name, _id_list, extra=None):
        self.db_name = db_name
        self.collection_name = collection_name
        self._id_list = _id_list
        self.extra = extra


class RefCallback(object):
    @classmethod
    def init_with_db(cls, db, query_path, set_path=None, query=None):
        return cls(db.db_name, db.collection_name, query_path, set_path)

    def __init__(self, db_name, collection_name, query_path, set_path=None, query=None):
        self.query = query
        self.query_path = query_path
        self.collection_name = collection_name
        self.db_name = db_name
        if set_path is None:
            set_path = query_path
        self.set_path = set_path


def simple_get_ref(db_name, collection_name, _id):
    if not isinstance(_id, bson.ObjectId):
        _id = bson.ObjectId(_id)
    return bson.dbref.DBRef(collection_name, _id, db_name)


async def get_ref(loader, callback=None):
    if isinstance(loader, RefLoader):
        return await _get_one_ref(loader, callback)
    elif isinstance(loader, RefListLoader):
        return await _get_many_ref(loader, callback)


def _get_one_ref(loader, callback):
    deref_db = DenormDBProxy()
    if callback is not None:
        deref_db.create(loader, callback)
    if loader.extra is not None:
        if loader.item is None:
            db = MongoDBAMQPProxy(db_name=loader.db_name, collection_name=loader.collection_name)
            item = db.get_by_id(loader._id)
        else:
            item = loader.item
        extra = deref_db.prepare_extra(item, loader.extra)
        item = bson.dbref.DBRef(loader.collection_name, loader._id, loader.db_name, _extra=extra)
    else:
        item = bson.dbref.DBRef(loader.collection_name, loader._id, loader.db_name)
    return item


async def _get_many_ref(loader, callback):
    deref_db = DenormDBProxy()
    if callback is not None:
        deref_db.create(loader, callback)
    if loader.extra is not None:
        db = MongoDBAMQPProxy(db_name=loader.db_name, collection_name=loader.collection_name)
        items = []
        for item in await db.find_by_id_list(loader._id_list):
            extra = deref_db.prepare_extra(item, loader.extra)
            items.append(bson.dbref.DBRef(loader.collection_name, item.get('_id'), loader.db_name, _extra=extra))
    else:
        items = [bson.dbref.DBRef(loader.collection_name, i, loader.db_name) for i in loader._id_list]
    return items



from aiohttp.web import BaseRequest

from asynclib.amqp.client import AMQPClient
from asynclib.http.error import BaseError


def parse_token(request, token_key='auth_token'):
    """

    :type request: BaseRequest
    """
    auth_token = None
    auth_header = request.headers.get('Authorization')
    if auth_header is not None and auth_header.lower().startswith('bearer '):
        auth_token = auth_header.split(' ')[1].strip()
    elif request.query.get(token_key):
        auth_token = request.query.get(token_key)
    elif request.cookies.get(token_key):
        auth_token = request.cookies.get(token_key)
    return auth_token


async def parse_user(request, service, endpoint, token_key='auth_token'):
    """
    :type token_key: str
    :type endpoint: str
    :type service: str
    :type request: BaseRequest
    """
    auth_token = parse_token(request, token_key)
    if auth_token is not None:
        try:
            async with AMQPClient(service) as client:
                user = await client.call(endpoint, **{token_key: auth_token})
            return user
        except BaseError:
            return None
    else:
        return None




class MetaSingleton(type):
    _instances = {}

    def __call__(cls, *args):
        if cls not in cls._instances:
            cls._instances[cls] = super(MetaSingleton, cls).__call__(*args)
        return cls._instances[cls]


from setuptools import setup, find_packages


setup(
    name='asynchttplib',
    version='0.3.3',
    description='Lib for creating microservice on asyncio',
    author='Oleksii Moisieienko',
    author_email='me@alexmay23.com',
    license='MIT',
    packages=find_packages(exclude=['dist','docs', 'tests']),
    install_requires=['aiohttp', 'aioamqp', 'motor', 'webargs', 'uvloop', 'aioredis', 'flatdict']
)


import asyncio
import json
import re

import aioamqp

from asynclib.amqp import logger
from asynclib.http.error import BaseError
from asynclib.utils.loop import get_event_loop


class JSONDumpError(BaseError):

    def __init__(self):
        super(JSONDumpError, self).__init__(
            code='JSON',
            description='JSON_DUMP_ERROR'
        )


class AMQPServer(object):
    def __init__(self, name='', prefix='am', host='localhost', user='guest', password='guest',
                 vhost='/', exchange_type='topic', dumper=json, prefetch_count=1):
        if not prefix.endswith('_'):
            prefix += '_'
        if len(name) > 0 and not name.endswith('_'):
            name += '_'
        self.credentials = {
            'host': host,
            'login': user,
            'password': password,
            'virtual_host': vhost
        }
        self.dumper = dumper
        self.prefix = prefix
        self.prefetch_count = prefetch_count
        self.name = name
        self.exchange_name = self.prefix + 'exchange_' + exchange_type
        self.queue_name = self.prefix + name + 'queue_' + exchange_type
        self.transport = None
        self.protocol = None
        self.channel = None
        self.exchange_type = exchange_type
        self.endpoints = []

    def register_endpoint(self, key, endpoint):
        key = (self.prefix + self.name).replace('_', '.') + key
        pattern = '[^.]*'.join([re.escape(i) for i in key.split('*')])
        self.endpoints.append((key, re.compile(r'^{}$'.format(pattern)), endpoint))

    async def handle(self, channel, body, envelope, properties):
        for key, pattern, endpoint in self.endpoints:
            if pattern.match(envelope.routing_key):
                try:
                    message = self.dumper.loads(body)
                    result = {'result': await endpoint(**message.get('parameters', {}))}
                    if self.exchange_type != 'fanout':
                        await channel.basic_client_ack(envelope.delivery_tag)
                except BaseError as e:
                    result = {'error': e.map()}
                logger.debug('finished execution')
                if properties.reply_to is not None:
                    try:
                        body = self.dumper.dumps(result)
                    except Exception:
                        body = self.dumper.dumps({'error': JSONDumpError().map()})
                    await self.channel.basic_publish(body, exchange_name='',
                                                     routing_key=properties.reply_to,
                                                     properties={
                                                         'correlation_id': properties.correlation_id
                                                     })

    async def prepare_queues(self):
        for key, pattern, endpoint in self.endpoints:
            await self.channel.queue_bind(self.queue_name, self.exchange_name, routing_key=key)
            if self.exchange_type == 'fanout':
                await self.channel.basic_consume(self.handle, queue_name=self.queue_name, no_ack=True)
            else:
                await self.channel.basic_consume(self.handle, queue_name=self.queue_name)
            logger.debug('Endpoint %s bound to %s', key, self.queue_name)

    async def connect(self):
        self.transport, self.protocol = await aioamqp.connect(**self.credentials)
        self.channel = await self.protocol.channel()
        await self.channel.exchange_declare(self.exchange_name, self.exchange_type)
        if self.exchange_type == 'fanout':
            self.queue_name = await self.channel.queue_declare()['queue']
        else:
            await self.channel.queue_declare(self.queue_name)
        await self.channel.basic_qos(0, prefetch_count=self.prefetch_count)
        await self.prepare_queues()

    def run(self, loop):
        loop.run_until_complete(self.start())
        print("App has started")
        loop.run_forever()

    async def start(self):
        try:
            await self.connect()
        except Exception as e:
            await self.terminate()
            print('Reconnecting, in case of', e)
            await asyncio.sleep(5)
            await self.start()

    async def terminate(self):
        if self.protocol is not None:
            await self.protocol.close()
            self.transport.close()



# coding=utf-8
import functools
from bson import ObjectId
from bson.errors import InvalidId
from marshmallow import fields, validate, Schema
from webargs import argmap2schema
import aiohttp.web

from asynclib.http.error import BaseError

fields = fields
validate = validate
Schema = Schema


class ErrorBodySchema(Schema):
    description = fields.String()
    code = fields.String()
    body = fields.Dict()


async def error_middleware(app, handler):

    async def middleware_handler(request):
        try:
            response = await handler(request)
            return response
        except BaseError as error:
            body = ErrorBodySchema().dumps(error.map()).data
            return aiohttp.web.Response(body=body, status=error.status, content_type='application/json')

    return middleware_handler


def dump(schema, data):
    body = schema().dumps(data).data
    return aiohttp.web.json_response(body=body, status=200, content_type='application/json')


class Nested(fields.Nested):
    """
    Кастомное поле Nested, для использования с множественными схемами
    Если указан callback - nested должен быть словарем
    callback принимает в качестве аргумента результат и возвращает ключ словаря nested (какую схему использовать)
    """

    def __init__(self, nested, callback=None, *args, **kwargs):
        self.callback = None
        if callback is not None:
            self.callback = callback
            nested = {a: self.__argmap2schema(b) for a, b in nested.items()}
            self._nested = nested
        else:
            nested = self.__argmap2schema(nested)
        super(Nested, self).__init__(nested, *args, **kwargs)

    def __argmap2schema(self, schema):
        if isinstance(schema, dict):
            schema = argmap2schema(schema)
        return schema

    def _serialize(self, nested_obj, attr, obj):
        if self.callback is not None:
            self.__schema = None
            self.nested = self._nested[self.callback(nested_obj)]
        result = super(Nested, self)._serialize(nested_obj, attr, obj)
        if self.callback is not None:
            self.nested = self._nested
        return result


class MongoId(fields.String):
    def _serialize(self, value, attr, obj):
        return super(MongoId, self)._serialize(value, attr, obj)

    def _deserialize(self, value, attr, data):
        value = super(MongoId, self)._deserialize(value, attr, data)
        try:
            return ObjectId(value)
        except InvalidId:
            self.fail('invalid')


class DateTimeReplaced(fields.DateTime):
    def _serialize(self, value, attr, obj):
        value = value.replace(microsecond=0) if value is not None else None
        return super(DateTimeReplaced, self)._serialize(value, attr, obj)

    def _deserialize(self, value, attr, obj):
        return super(DateTimeReplaced, self)._serialize(value, attr, obj)


class OKSchema(Schema):
    ok = fields.Boolean()


def make_list_schema(schema, *args, **kwargs):
    class _Shema(Schema):
        total = fields.Integer()
        objects = fields.List(Nested(schema, *args, **kwargs))

    return _Shema


skip_limit_args = {'skip': fields.Integer(), 'limit': fields.Integer()}

fields.MongoId = MongoId